#include "argument"
#include <crdb.h>
#include <cstdlib> // for getenv
#include <fstream>
#include <iostream>
#include <string>

namespace crdb_bin {
std::string toupper(std::string str) {
    for (auto& c : str) {
        c = std::toupper(c);
    }
    return str;
}
std::string tolower(std::string str) {
    for (auto& c : str) {
        c = std::tolower(c);
    }
    return str;
}
}; // namespace crdb_bin

int main(int argc, char* argv[]) {

    std::string cache_dir;
    if (const char* env_p = std::getenv("CRDB_CACHE_DIR")) {
        cache_dir = env_p;
    } else {
        cache_dir = "/tmp/crdb_cache";
    }

    yuc::argument arg;

    arg.add_option('d', "data", "use input data file", "");
    arg.add_option('t', "etype", "energy type [EKN],EK,R,ETOT,ETOTN", "");
    arg.add_option('A', "accuracy", "conversion accuracy 0-2", "1");
    arg.add_option('s', "subexp",
                   "name (and time intervals) of sub-experiments", "");
    arg.add_option('o', "output", "output filename", "");
    arg.add_flag('a', "append", "append to output file", false);

    arg.add_usage("[-t ETYPE] [-s SUBEXP] [-o FILE] <numerator> [denominator]");

    _YUC_ARGUMENT_PARSE_ASSERTION(arg, argc, argv);

    std::string energy_type = crdb_bin::toupper(arg["etype"].str());
    if (energy_type.empty() && !(arg["data"])) {
        energy_type = "EKN";
    }
    std::string subexp      = arg["subexp"];
    std::string ofn         = arg["output"];
    if (ofn == "-") {
        ofn = "/dev/stdout";
    }

    std::string num = "", den = "";
    switch (arg.size()) {
    case 3:
        den = arg[2];
    case 2:
        num = arg[1];
    case 1:
        break;
    default:
        std::cerr << arg[0] << ": too many CR quantities provied" << std::endl;
        arg.help();
        return EXIT_FAILURE;
    }

    std::string y_quan  = ""; // for filter datafile
    std::string acc_str = arg["accuracy"];
    auto acc            = CRRequest::exact;
    switch (acc_str.size() == 1 ? acc_str[0] : ' ') {
    case '0':
        acc = CRRequest::none;
        break;
    case '1':
        acc = CRRequest::exact;
        break;
    case '2':
        acc = CRRequest::approx;
        break;
    default:
        std::cerr << arg[0] << ": accuracy should be 0, 1 or 2" << std::endl;
        return EXIT_FAILURE;
    }
    CRRequest req{
        .num = num, .den = den, .energy_type = energy_type, .accuracy = acc};
    if (arg["data"]) {
        if (num.size()) {
            y_quan = num;
            if (den.size()) {
                y_quan += "/" + den;
            }
        }
    } else { // fetch online data or use cached data, must provide cr name
        if (num.empty()) {
            std::cerr << arg[0] << " must provicde numerator [denominator]"
                      << std::endl;
        }
    }

    if (ofn.empty()) { // list experiments
        CRDataset dataset;
        if (arg["data"]) {
            dataset.parse_file(arg["data"].str());
        } else {
            dataset = req.fetch(cache_dir);
        }
        size_t len_name = 0, len_exp = 0;
        std::vector<size_t> idxs;
        for (size_t i = 0; i < dataset.size(); ++i) {
            const auto& d = *dataset[i];
            if (arg["data"]) {
                if ((!energy_type.empty() &&
                     energy_type != crdb_bin::toupper(d.x_quantity)) ||
                    (y_quan.size() && y_quan != d.y_quantity)) {
                    continue;
                }
            }
            if (subexp.empty() ||
                d.experiment.substr(0, subexp.size()) == subexp) {
                if (len_name < d.y_quantity.size()) {
                    len_name = d.y_quantity.size();
                }
                if (len_exp < d.experiment.size()) {
                    len_exp = d.experiment.size();
                }
                idxs.push_back(i);
            }
        }
        for (auto i : idxs) {
            const auto& d = *dataset[i];
            std::printf(("%-" + std::to_string(len_name + 3) +
                         "s   %s    %.1e ~ %.1e   %2ld    %-" +
                         std::to_string(len_exp + 3) + "s  %s\n")
                            .c_str(), //
                        d.y_quantity.c_str(), d.x_quantity.c_str(), d.x().min(),
                        d.x().max(), d.size(), d.experiment.c_str(),
                        d.cite.c_str());
        }
    } else {
        // save expriments
        std::stringstream ss;
        if (arg["data"]) {
            ss << std::ifstream(arg["data"].str()).rdbuf();
        } else {
            req.fetch(ss, cache_dir);
        }
        std::string line;
        bool appending = arg["append"];
        std::ofstream ofs(ofn, appending ? (ofs.out | ofs.app) : ofs.out);
        while (std::getline(ss, line)) {
            std::stringstream ssl(line);
            std::string field;
            ssl >> field;
            if (field.size() && ((field[0] == '#') || subexp.empty() ||
                                 ((ssl >> field) &&
                                  field.substr(0, subexp.size()) == subexp))) {
                ofs << line << std::endl;
            }
        }
    }

    return 0;
}
