#define WITHOUT_ROOT
#include "crinfo.hh"
#include "crdb.h"

std::string CRData::particle_number::str(void) const {
    return "(" + (Z == any ? "any" : std::to_string(Z)) + //
           "," + (A == any ? "any" : std::to_string(A)) + ")";
}

constexpr auto& any = CRData::particle_number::any;

bool particle_in_list(int Z, int A, //
                      const std::vector<CRData::particle_number>& lst) {
    for (const auto& n : lst) {
        if (n.Z == any || n.Z == Z) {
            if (n.A == any || n.A == A) {
                return true;
            }
        }
    }
    return false;
}
bool CRData::num_has(int Z, int A) const {
    return particle_in_list(Z, A, num_list);
}
bool CRData::den_has(int Z, int A) const {
    return particle_in_list(Z, A, den_list);
}
bool CRData::has(int Z, int A) const {
    return num_has(Z, A) || is_ratio && den_has(Z, A);
}

void canonicalize_num_den(std::string& num, std::string& den) {
    // "A/B" "" => "A" "B"
    size_t pos_slash = num.find('/');
    if (pos_slash != num.npos) {
        den = num.substr(pos_slash + 1);
        num = num.substr(0, pos_slash);
    }
}

std::vector<std::string> cr_split(const std::string& s) {
    // "A+B+e++C" => {"A", "B", "e+", "C"}
    std::vector<std::string> res;
    size_t bg = 0, ed = 0;
    while (s.npos != (ed = s.find('+', bg))) {
        if (bg + 1 == ed && s[bg] == 'e') { // "e+"
            ++ed;
            if (ed == s.size()) {
                ed = s.npos;
                break;
            }
        }
        if (bg != ed) {
            res.push_back(s.substr(bg, ed - bg));
        }
        bg = ed + 1;
    }
    if (bg < s.size()) {
        res.push_back(s.substr(bg));
    }
    return res;
}

static std::map<std::string, CRData::particle_number> cr_dict = {
    {"All", {any, any}}, {"AllParticles", {any, any}}, //

    {"e-", {-1, 0}},     {"electron", {-1, 0}}, //
    {"e+", {1, 0}},      {"positron", {1, 0}},  //

    {"P", {1, 1}},       {"Proton", {1, 1}}, //

    {"H", {1, any}},     {"Hydrogen", {1, any}}, //
    {"D", {1, 2}},       {"Deuterium", {1, 2}},  //
    {"T", {1, 3}},       {"Tritium", {1, 3}},    //
    {"1H", {1, 1}},      {"2H", {1, 2}},
    {"3H", {1, 3}}, //

    {"He", {2, any}},    {"Helium", {2, any}}, //
    {"3He", {2, 3}},     {"4He", {2, 4}},      //
};

CRData::particle_number cr_ident(const std::string& name) {
    auto it = cr_dict.find(name);
    if (it != cr_dict.end()) {
        return it->second;
    }
    return {0, 0};
    // throw std::runtime_error("unrecognized CR name '" + name + "'");
}

bool get_info(CRData& data, const std::string& num, const std::string& den) {
    std::string num_str = num.empty() ? data.y_quantity : num;
    std::string den_str = den;
    canonicalize_num_den(num_str, den_str);

    data.is_ratio = !den_str.empty();

    data.num_list.clear();
    data.den_list.clear();
    data.known_species = false;

    for (const auto& name : cr_split(num_str)) {
        data.num_list.push_back(cr_ident(name));
        const auto& n = data.num_list.back();
        if (0 == n.A && 0 == n.Z) {
            data.num_list.clear();
            return false;
        }
    }
    if (data.is_ratio) {
        for (const auto& name : cr_split(den_str)) {
            data.den_list.push_back(cr_ident(name));
            const auto& n = data.den_list.back();
            if (0 == n.A && 0 == n.Z) {
                data.num_list.clear();
                data.den_list.clear();
                return false;
            }
        }
    }
    data.known_species = true;
    return true;
}
