#include "crutil.hh"
#include "loginterp"
#include <crdb.h>
#include <iostream>

#include <array>
#include <cmath>
#include <fstream>
#include <sstream>
#include <vector>

#ifndef WITHOUT_ROOT
#include <TColor.h>
#include <TError.h>
#include <TFile.h>
#include <TGaxis.h>
#include <TGraphAsymmErrors.h>
#include <TH1D.h>
#include <TLegend.h>
#include <TPad.h>
int gErrorIgnoreLevel = kWarning;
#endif

namespace _CRPLOT {
template <size_t N>
std::array<double, N> line_columns(const std::string& line,
                                   const std::array<size_t, N>& cols = {0}) {
    std::stringstream ss(line);
    std::array<double, N> ret;

    size_t col  = 0; // column order in file
    size_t read = 0; // number of quested columns read
    double val;

    if (cols[0] == 0) {
        while (col < N && ss >> ret[col++]) {
            ++read;
        }
    } else {
        while (ss >> val) {
            ++col;
            for (size_t i = 0; i < N; ++i) {
                if (cols[i] == col) {
                    ret[i] = val;
                    ++read;
                }
            }
        }
    }

    if (read < N) { // not all quest have been fulfilled
        ret[0] = NAN;
    }

    return ret;
}

template <size_t N>
std::array<std::vector<double>, N>
read_columns(const std::string& filename,
             const std::array<size_t, N>& cols = {0}) {

    _CRUTIL::FileParser fp(filename);
    std::array<std::vector<double>, N> ret;
    while (fp.next_line(fp.skip::both)) {
        auto vals = line_columns(fp.current_line(), cols);
        if (std::isnan(vals[0])) {
            fp.panic("failed to parse columns");
        }
        for (size_t i = 0; i < N; ++i) {
            ret[i].push_back(vals[i]);
        }
    }
    return ret;
}

}; // namespace _CRPLOT

CRPlot::CRPlot(const std::string& title, int canv_w, int canv_h)
    : title(title),                    //
      x_min(NAN), x_max(NAN),          //
      y_min(NAN), y_max(NAN),          //
      log_x(true), log_y(true),        //
      grid_x(true), grid_y(true),      //
      scale_index(0), legend_phase(4), //
      ratio_min(NAN), ratio_max(NAN),  //
      canvas_width(canv_w),            //
      canvas_height(canv_h),           //
      divide_ratio(0.3) {}

using Style = CRPlot::Style;
void CRPlot::AddData(const CRData& data, const Style& style,
                     const std::string& label) {
    std::string l = label.empty() ? data.label : label;
    _egraphs.erase(l);
    _egraphs.insert({l, {data, style}});
}
// void CRPlot::AddData(const std::string& filename, const std::string& label,
//                      const std::array<size_t, 3> columns) {
//     std::string l = label.empty() ? data->label : label;
//     const auto ret = _CRPLOT::read_columns(filename, columns);
// }
void CRPlot::DropData(const std::string& label) {
    if (label.empty()) {
        _egraphs.clear();
    } else {
        _egraphs.erase(label);
    }
}

void CRPlot::AddCurve(const CRCurve& curve, const Style& style,
                      const std::string& label) {
    _graphs.erase(label);
    _graphs.insert({label, {curve, style}});
}
void CRPlot::AddCurve(const std::string& filename, const Style& style,
                      const std::string& label,
                      const std::array<size_t, 2>& columns) {
    std::string l  = label.empty() ? filename : label;
    const auto ret = _CRPLOT::read_columns(filename, columns);
    const size_t n = ret[0].size();
    AddCurve({std::valarray<double>(ret[0].data(), n),
              std::valarray<double>(ret[1].data(), n)},
             style, l);
}
void CRPlot::DropCurve(const std::string& label) {
    if (label.empty()) {
        _graphs.clear();
    } else {
        _graphs.erase(label);
    }
}

void CRPlot::AddBand(const CRBand& band, const Style& style,
                     const std::string& label) {
    _bands.erase(label);
    _bands.insert({label, {band, style}});
}
void CRPlot::AddBand(const std::string& filename, const Style& style,
                     const std::string& label,
                     const std::array<size_t, 4>& columns) {
    std::string l  = label.empty() ? filename : label;
    const auto ret = _CRPLOT::read_columns(filename, columns);
    const size_t n = ret[0].size();
    AddBand({std::valarray<double>(ret[0].data(), n),
             std::valarray<double>(ret[1].data(), n),
             std::valarray<double>(ret[2].data(), n),
             std::valarray<double>(ret[3].data(), n)},
            style, l);
}
void CRPlot::DropBand(const std::string& label) {
    if (label.empty()) {
        _bands.clear();
    } else {
        _bands.erase(label);
    }
}

double AdjustXMin(const CRPlot& plot) {
    double x = NAN;
    for (const auto& d : plot._egraphs) {
        double xmin = d.second.first.x_min();
        if (!(x <= xmin)) { // including nan
            x = xmin;
        }
    }
    return x;
}
double AdjustXMax(const CRPlot& plot) {
    double x = NAN;
    for (const auto& d : plot._egraphs) {
        double xmax = d.second.first.x_max();
        if (!(x >= xmax)) { // including nan
            x = xmax;
        }
    }
    return x;
}
double AdjustYMin(const CRPlot& plot) {
    double y = NAN;
    for (const auto& d : plot._egraphs) {
        for (const auto& dp : d.second.first) {
            double scaled_y = std::pow(dp.x, plot.scale_index) *
                              (plot.log_y ? std::pow(dp.y - dp.ey1, 2) / dp.y
                                          : (dp.y - 2 * dp.ey1));
            if (!(y <= scaled_y)) { // including nan
                y = scaled_y;
            }
        }
    }
    return y;
}
double AdjustYMax(const CRPlot& plot) {
    double y = NAN;
    for (const auto& d : plot._egraphs) {
        for (const auto& dp : d.second.first) {
            double scaled_y = std::pow(dp.x, plot.scale_index) *
                              (plot.log_y ? std::pow(dp.y + dp.ey2, 2) / dp.y
                                          : (dp.y + 2 * dp.ey2));
            if (!(y >= scaled_y)) { // including nan
                y = scaled_y;
            }
        }
    }
    return y;
}

#ifndef WITHOUT_ROOT
std::shared_ptr<TCanvas> CRPlot::Plot(void) const {
    const bool with_ratio = !ratio_ref.empty();
    double x_min_plot     = std::isnan(x_min) ? AdjustXMin(*this) : x_min;
    double x_max_plot     = std::isnan(x_max) ? AdjustXMax(*this) : x_max;
    double y_min_plot     = std::isnan(y_min) ? AdjustYMin(*this) : y_min;
    double y_max_plot     = std::isnan(y_max) ? AdjustYMax(*this) : y_max;
    double y_min_ratio = with_ratio && std::isnan(ratio_min) ? 0.5 : ratio_min;
    double y_max_ratio =
        with_ratio && std::isnan(ratio_max) ? (log_y ? 2.0 : 1.5) : ratio_max;
    if (std::isnan(x_min_plot) || std::isnan(x_max_plot) ||
        std::isnan(y_min_plot) || std::isnan(y_max_plot)) {
        throw std::runtime_error("invalid plot range");
    }

    // calclulate canvas layout
    const double label_size = canvas_width * 0.02;
    const double title_size = canvas_width * 0.03;

    const int label_font = 133;
    const int title_font = 133;

    // overall canvas
    auto pcanv =
        std::make_shared<TCanvas>("canv", "canv", canvas_width, canvas_height);
    auto& canv = *pcanv;
    // canv.SetFillColor(0);
    // canv.SetGrid(grid_x, grid_y);
    canv.SetBorderMode(0);
    canv.SetFrameBorderMode(0);

    // th1d for set plot range range
    TH1D *phframe = nullptr, *phframe_ratio = nullptr;

    phframe = new TH1D("", title.c_str(), 1, x_min_plot, x_max_plot);
    phframe->SetMinimum(y_min_plot);
    phframe->SetMaximum(y_max_plot);
    if (with_ratio) {
        phframe_ratio = new TH1D("", "", 1, x_min_plot, x_max_plot);
        phframe_ratio->SetMinimum(y_min_ratio);
        phframe_ratio->SetMaximum(y_max_ratio);
    }

    for (auto phf : {phframe, phframe_ratio}) {
        if (!phf) {
            continue;
        }
        phf->SetBit(kCanDelete); // prevent leak
        phf->SetStats(0);
        phf->GetYaxis()->SetDecimals(!log_y);
        phf->GetYaxis()->SetLabelFont(label_font);
        phf->GetYaxis()->SetLabelSize(label_size);
    }

    phframe->GetYaxis()->SetTitleFont(title_font);
    phframe->GetYaxis()->SetTitleSize(title_size);
    phframe->GetYaxis()->SetTitle(y_label.c_str());

    auto xax = (phframe_ratio ? phframe_ratio : phframe)->GetXaxis();
    xax->SetTitleSize(title_size);
    xax->SetTitleFont(title_font);
    xax->SetTitle(x_label.c_str());
    xax->SetLabelFont(label_font);
    xax->SetLabelSize(label_size);
    xax->SetTitleOffset(3.);
    xax->SetDecimals(!log_x);

    TPad *pad_plot = nullptr, *pad_ratio = nullptr;
    yuc::log_interpolator interp_fcn;
    if (with_ratio) {
        pad_plot  = new TPad("", "pad_plot", 0, divide_ratio, 1, 1);
        pad_ratio = new TPad("", "pad_ratio", 0, 0, 1, divide_ratio);

        pad_plot->SetBottomMargin(0);
        pad_ratio->SetTopMargin(0);
        pad_ratio->SetBottomMargin(0.1 / divide_ratio);

        // prepare interpolation function TODO: linear interp for linear data
        auto ref_graph  = _graphs.find(ratio_ref);
        auto ref_egraph = _egraphs.find(ratio_ref);
        auto ref_band   = _bands.find(ratio_ref);
        if (ref_graph != _graphs.end()) {
            interp_fcn.insert(ref_graph->second.first.x,
                              ref_graph->second.first.y);
        } else if (ref_egraph != _egraphs.end()) {
            interp_fcn.insert(ref_egraph->second.first.x(),
                              ref_egraph->second.first.y());
        } else if (ref_band != _bands.end()) {
            interp_fcn.insert(ref_band->second.first.x,
                              ref_band->second.first.y);
        } else {
            throw std::runtime_error("ratio reference not found: '" +
                                     ratio_ref + "'");
        }
    } else {
        pad_plot = new TPad("", "pad_plot", 0, 0, 1, 1);
    }
    for (auto spad : {pad_plot, pad_ratio}) {
        if (!spad) {
            continue;
        }
        spad->SetLogx(log_x);
        spad->SetLogy(log_y);
        spad->SetGrid(grid_x, grid_y);
        spad->SetFillColor(0);
        pcanv->cd();
        spad->Draw();
    }

    pad_plot->cd();
    phframe->Draw();
    if (with_ratio) {
        pad_ratio->cd();
        phframe_ratio->Draw();
    }

    // prepare legends
    size_t n_entry = 0;
    auto leg       = new TLegend(0.1, with_ratio ? 0 : 0.1, 0.9, 0.9);
    leg->SetBit(kCanDelete); // prevent leak

    // do not delete TGraph and TGraphAsymmErrors until finished!
    std::vector<TGraph*> graphs;
    std::vector<TGraphAsymmErrors*> egraphs;

    std::vector<TGraph*> ratio_graphs;
    std::vector<TGraphAsymmErrors*> ratio_egraphs;

    for (const auto& d : _bands) {
        const auto& band                = d.second.first;
        const auto style                = d.second.second;
        const std::string draw_flag     = style.size > 0 ? "EL3" : "E3";
        const std::string draw_leg_flag = style.size > 0 ? "fl" : "f";
        const size_t n                  = band.x.size();
        auto g                          = new TGraphAsymmErrors(n);
        auto rg = with_ratio ? new TGraphAsymmErrors(n) : nullptr;
        g->SetTitle(d.first.c_str());
        for (size_t i = 0; i < n; ++i) {
            double x       = band.x[i];
            double y       = band.y[i];
            double y_upper = band.y_upper[i];
            double y_lower = band.y_lower[i];
            if (y_upper < y_lower) { // swap
                y_upper += y_lower;
                y_lower = y_upper - y_lower;
                y_upper -= y_lower;
            }
            double el = y - y_lower;
            double eu = y_upper - y;

            double fac = std::pow(x, scale_index);

            g->SetPoint(i, x, y * fac);
            g->SetPointError(i, 0, 0, el * fac, eu * fac);
            if (rg) {
                double ref_y = interp_fcn(x);
                rg->SetPoint(i, x, y / ref_y);
                rg->SetPointError(i, 0, 0, el / ref_y, eu / ref_y);
            }
        }
        for (auto sg : {g, rg}) {
            if (!sg) {
                continue;
            }
            sg->SetBit(kCanDelete); // prevent leak
            if (style.fill_style / 1000 == 1) {
                double alpha = (style.fill_style % 1000 % 101) * .01;
                sg->SetFillColorAlpha(style.fill_color < 0 ? style.color
                                                           : style.fill_color,
                                      alpha);
                sg->SetFillStyle(1001);
            } else {
                sg->SetFillColor(style.fill_color < 0 ? style.color
                                                      : style.fill_color);
                sg->SetFillStyle(style.fill_style);
            }
            if (style.size > 0) {
                // draw center line
                sg->SetLineWidth(style.size);
                sg->SetLineColor(style.color);
                sg->SetLineStyle(style.style);
            } else {
                sg->SetLineWidth(0);
                sg->SetLineColor(0);
                sg->SetLineStyle(0);
            }
        }
        pad_plot->cd();
        g->Draw(("same " + draw_flag).c_str());
        egraphs.push_back(g);
        if (rg) {
            pad_ratio->cd();
            rg->Draw(("same " + draw_flag).c_str());
            ratio_egraphs.push_back(g);
        }
        leg->AddEntry(g, d.first.c_str(), draw_leg_flag.c_str());
        ++n_entry;
    }

    // plot curves
    for (const auto& d : _graphs) {
        const std::string draw_flag = "l";
        const auto& curve           = d.second.first;
        const auto style            = d.second.second;
        const size_t n              = curve.x.size();
        auto g                      = new TGraph(curve.x.size());
        auto rg                     = with_ratio ? new TGraph(n) : nullptr;
        g->SetTitle(d.first.c_str());
        for (size_t i = 0; i < curve.x.size(); ++i) {
            double x = curve.x[i], y = curve.y[i],
                   fac = std::pow(x, scale_index);
            g->SetPoint(i, x, y * fac);
            if (rg) {
                double ref_y = interp_fcn(x);
                rg->SetPoint(i, x, y / ref_y);
            }
        }
        for (auto sg : {g, rg}) {
            if (!sg) {
                continue;
            }
            sg->SetBit(kCanDelete); // prevent leak
            sg->SetLineColor(style.color);
            sg->SetLineStyle(style.style);
            sg->SetLineWidth(style.size < 0 ? 2 : style.size);
        }
        pad_plot->cd();
        g->Draw(("same " + draw_flag).c_str());
        graphs.push_back(g);
        if (rg) {
            pad_ratio->cd();
            rg->Draw(("same " + draw_flag).c_str());
            ratio_graphs.push_back(g);
        }
        leg->AddEntry(g, d.first.c_str(), draw_flag.c_str());
        ++n_entry;
    }

    // plot data
    for (const auto& d : _egraphs) {
        const std::string draw_flag = "ep";
        const auto data             = d.second.first;
        const auto style            = d.second.second;
        const size_t n              = data.size();
        auto g                      = new TGraphAsymmErrors(data.size());
        auto rg = with_ratio ? new TGraphAsymmErrors(n) : nullptr;
        g->SetTitle(d.first.c_str());
        for (size_t i = 0; i < data.size(); ++i) {
            auto dp    = data[i];
            double fac = std::pow(dp.x, scale_index);

            // TODO: support upper-limit

            g->SetPoint(i, dp.x, dp.y * fac);
            g->SetPointError(i,                 //
                             dp.x - dp.x_lower, //
                             dp.x_upper - dp.x, //
                             dp.ey1 * fac, dp.ey2 * fac);
            if (rg) {
                double ref_y = interp_fcn(dp.x);
                rg->SetPoint(i, dp.x, dp.y / ref_y);
                rg->SetPointError(i, dp.x - dp.x_lower, dp.x_upper - dp.x,
                                  dp.ey1 / ref_y, dp.ey2 / ref_y);
            }
        }
        for (auto sg : {g, rg}) {
            if (!sg) {
                continue;
            }
            sg->SetBit(kCanDelete); // prevent leak
            sg->SetFillColor(0);
            sg->SetLineColor(style.color);
            sg->SetMarkerColor(style.color);
            sg->SetMarkerStyle(style.style);
            sg->SetLineWidth(style.size < 2 ? 1 : (style.size - 1));
            sg->SetMarkerSize(style.size < 0 ? 2 : style.size);
        }
        pad_plot->cd();
        g->Draw(("same " + draw_flag).c_str());
        egraphs.push_back(g);
        if (rg) {
            pad_ratio->cd();
            rg->Draw(("same " + draw_flag).c_str());
            ratio_egraphs.push_back(g);
        }
        leg->AddEntry(g, d.first.c_str(), draw_flag.c_str());
        ++n_entry;
    }

    switch (legend_phase) {
    case 1:
        leg->SetX1(0.6);
        leg->SetY1(leg->GetY2() - n_entry * 0.03);
        break;
    case 2:
        leg->SetX2(0.4);
        leg->SetY1(leg->GetY2() - n_entry * 0.03);
        break;
    case 3:
        leg->SetX2(0.4);
        leg->SetY2(leg->GetY1() + n_entry * 0.03);
        break;
    case 4:
        leg->SetX1(0.6);
        leg->SetY2(leg->GetY1() + n_entry * 0.03);
        break;
    }
    // TODO multi-column legend

    pad_plot->cd();
    leg->Draw();

    return pcanv;

    // TODO: test memory leak!
    // canv.SaveAs(filename.c_str());

    // for (auto p : egraphs) {
    //     delete p;
    // }
    // for (auto p : graphs) {
    //     delete p;
    // }
}
#endif
