#pragma once

#include <fstream>
#include <stdexcept>
#include <string>
#include <algorithm>

namespace _CRUTIL {

// equivalent 'strncasecmp'/'strnicmp'
int strncasecmp(const char* s1, const char* s2, size_t n);
bool is_substr_i(const std::string& sup, const std::string& sub, int pos = 0);

// for parsing file
// struct ParseState {
//     size_t line_number;
//     std::string filename;
//     std::string line;
//     bool getline(std::istream& is);
//     template <typename Ex = std::runtime_error>
//     void panic(const std::string& msg = "unexpected input") const {
//         throw Ex(filename + ":" + std::to_string(line_number) + ": " + msg);
//     }
// };

class StreamParser {
  public:
    enum skip : uint8_t {
        none    = 0x00,
        empty   = 0x01,
        hashtag = 0x02,
        both    = 0x03
    };

  protected:
    std::string _line, _filename;
    size_t _line_number;

  public:
    StreamParser(std::string filename = "[stream]") : _line(""), _filename(filename), _line_number(0) {}

  public:
    const std::string& current_line(void) const { return _line; }
    const std::string& filename(void) const { return _filename; }
    size_t current_line_number(void) const { return _line_number; }

  public:
    template <typename Ex = std::runtime_error>
    void panic(const std::string& msg = "unexpected input") const {
        throw Ex(_filename + ":" + std::to_string(_line_number) + ": " + msg);
    }

  public:
    StreamParser& next_line(std::istream& is, skip s = skip::none) {
        while (std::getline(is, _line)) {
            ++_line_number;
            // trim line
            _line.erase(
                _line.begin(),
                std::find_if(_line.begin(), _line.end(), [](unsigned char ch) {
                    return !std::isspace(ch);
                }));
            _line.erase(
                std::find_if(_line.rbegin(), _line.rend(),
                             [](unsigned char ch) { return !std::isspace(ch); })
                    .base(),
                _line.end());

            if (_line.empty() ? !(s & skip::empty)
                              : _line[0] != '#' || !(s & skip::hashtag)) {
                break;
            }
        }
        return *this;
    }
};

class FileParser : protected std::ifstream {
  public:
    enum skip : uint8_t {
        none    = 0x00,
        empty   = 0x01,
        hashtag = 0x02,
        both    = 0x03
    };

  public:
    using std::ifstream::operator bool;
    using std::ifstream::good;

  protected:
    std::string _line, _filename;
    size_t _line_number;

  public:
    FileParser(const std::string& filename);

  public:
    const std::string& current_line(void) const { return _line; }
    const std::string& filename(void) const { return _filename; }
    size_t current_line_number(void) const { return _line_number; }

  public:
    FileParser& next_line(skip s = skip::none);

  public:
    template <typename Ex = std::runtime_error>
    void panic(const std::string& msg = "unexpected input") const {
        throw Ex(_filename + ":" + std::to_string(_line_number) + ": " + msg);
    }
};

}; // namespace _CRUTIL
